/*
 * vector.hpp
 *
 * Created on: Feb 28, 2014
 * Author: Jan Travnicek
 */

#ifndef __VECTOR_HPP_
#define __VECTOR_HPP_

namespace std {

template< class T , class ... Ts >
std::ostream& operator<<(std::ostream& out, const std::vector<T, Ts ...>& vector) {
	out << "[";

	bool first = true;
	for(const T& item : vector) {
		if(!first) out << ", ";
		first = false;
		out << item;
	}

	out << "]";
	return out;
}

template<class T, class ... Ts>
struct compare<vector<T, Ts ...>> {
	int operator()(const vector<T, Ts...>& first, const vector<T, Ts...>& second) const {
		if(first.size() < second.size()) return -1;
		if(first.size() > second.size()) return 1;

		static compare<typename std::decay < T >::type > comp;
		for(auto iterF = first.begin(), iterS = second.begin(); iterF != first.end(); ++iterF, ++iterS) {
			int res = comp(*iterF, *iterS);
			if(res != 0) return res;
		}
		return 0;
	}
};

template < class T, class ... Ts >
string to_string ( const std::vector < T, Ts ... > & value ) {
	std::stringstream ss;
	ss << value;
	return ss.str();
}

template < class ... Ts >
vector < bool, Ts ... > & operator |= ( vector < bool, Ts ... > & A, const vector < bool, Ts ... > & B ) {
	A.resize ( std::max ( A.size ( ), B.size ( ) ) );

	typename vector < bool, Ts ... >::iterator itA = A.begin ( );
	typename vector < bool, Ts ... >::const_iterator itB = B.begin ( );

	// c++ implementation-specific
	while ( itB < B.end ( ) ) // A is longer or of the same size as B
		* ( itA._M_p ++ ) |= * ( itB._M_p ++ ); // word-at-a-time bitwise operation

	// The rest of A above the size of B can be left intact

	return A;
}

template < class ... Ts >
vector < bool, Ts ... > operator | ( vector < bool, Ts ... > A, const vector < bool, Ts ... > & B ) {
	A |= B;
	return A;
}

template < class ... Ts >
vector < bool, Ts ... > & operator &= ( vector < bool, Ts ... > & A, const vector < bool, Ts ... > & B ) {
	A.resize ( std::max ( A.size ( ), B.size ( ) ) );

	typename vector < bool, Ts ... >::iterator itA = A.begin ( );
	typename vector < bool, Ts ... >::const_iterator itB = B.begin ( );

	// c++ implementation-specific
	while ( itB < B.end ( ) ) // A is longer or of the same size as B
		* ( itA._M_p ++ ) &= * ( itB._M_p ++ ); // word-at-a-time bitwise operation

	while ( itA < A.end ( ) ) // The rest of A above the size of B shall be zeroed
		* ( itA._M_p ++ ) = 0;

	return A;
}

template < class ... Ts >
vector < bool, Ts ... > operator & ( vector < bool, Ts ... > A, const vector < bool, Ts ... > & B ) {
	A &= B;
	return A;
}

template < class ... Ts >
vector < bool, Ts ... > & operator ^= ( vector < bool, Ts ... > & A, const vector < bool, Ts ... > & B ) {
	A.resize ( std::max ( A.size ( ), B.size ( ) ) );

	typename vector < bool, Ts ... >::iterator itA = A.begin ( );
	typename vector < bool, Ts ... >::const_iterator itB = B.begin ( );

	// c++ implementation-specific
	while ( itB < B.end ( ) ) // A is longer or of the same size as B
		* ( itA._M_p ++ ) ^= * ( itB._M_p ++ ); // word-at-a-time bitwise operation

	while ( itA < A.end ( ) ) // The rest of A above the size of B shall be flipped
		* ( itA._M_p ++ ) = ~ * ( itA._M_p ++ );

	return A;
}

template < class ... Ts >
vector < bool, Ts ... > operator ^ ( vector < bool, Ts ... > A, const vector < bool, Ts ... > & B ) {
	A ^= B;
	return A;
}

template < class ... Ts >
vector < bool, Ts ... > operator ~ ( vector < bool, Ts ... > A ) {
	A.flip ( );
	return A;
}

// the same type as the vector bool's internal storage type
typedef typename decay < decltype ( * ( declval < vector < bool > > ( ).begin ( )._M_p ) ) >::type vectorBoolInternalType;
// the size of the vector bool's internal storage type
const unsigned vectorBoolInternalTypeInBits = sizeof ( vectorBoolInternalType ) * 8;

// private helper for mask computation
inline vectorBoolInternalType getMask ( size_t dist ) {
	return ( ( vectorBoolInternalType { } + 1 ) << dist ) - 1;
}

template < class ... Ts >
vector < bool, Ts ... > & operator <<= ( vector < bool, Ts ... > & A, size_t dist ) {
	if ( A.size ( ) == 0 )
		return A;

	size_t distBlocks = dist / vectorBoolInternalTypeInBits;
	size_t distWithin = dist % vectorBoolInternalTypeInBits;
	size_t backDist = vectorBoolInternalTypeInBits - distWithin;

	// shift blocks if needed
	if ( distBlocks ) {
		// simulate behavior of reverse iterator
		auto itAReverse = A.end ( ) - 1;
		auto itASource = itAReverse;
		itASource._M_p -= distBlocks;

		while ( itASource >= A.begin ( ) )
			* ( itAReverse._M_p -- ) = * ( itASource._M_p -- );

		while ( itAReverse >= A.begin ( ) )
			* ( itAReverse._M_p -- ) = 0;
	}

	if ( distWithin == 0 )
		return A;

	// shift by the rest dist
	{
		vectorBoolInternalType bits1 { };
		vectorBoolInternalType bits2 { };

		// it might be more clear to iterate from the begining. However this is working nicely
		auto itA = A.begin ( );

		while ( itA < A.end ( ) ) {
			bits2 = * ( itA._M_p );
			* ( itA._M_p ) = * ( itA._M_p ) << distWithin | bits1 >> backDist;
			bits1 = bits2;

			itA._M_p ++;
		}
	}

	return A;
}

template < class ... Ts >
vector < bool, Ts ... > operator << ( vector < bool, Ts ... > A, size_t dist ) {
	A <<= dist;
	return A;
}

template < class ... Ts >
vector < bool, Ts ... > & operator >>= ( vector < bool, Ts ... > & A, size_t dist ) {
	if ( A.size ( ) == 0 )
		return A;

	size_t distBlocks = dist / vectorBoolInternalTypeInBits;
	size_t distWithin = dist % vectorBoolInternalTypeInBits;
	size_t sizeWithin = A.size ( ) % vectorBoolInternalTypeInBits;
	size_t backDist = vectorBoolInternalTypeInBits - distWithin;

	// shift blocks if needed
	if ( distBlocks ) {
		auto itA = A.begin ( );
		auto itASource = itA;
		itASource._M_p += distBlocks;

		while ( itASource < A.end ( ) )
			* ( itA._M_p ++ ) = * ( itASource._M_p ++ );

		while ( itA < A.end ( ) )
			* ( itA._M_p ++ ) = 0;
	}

	if ( distWithin == 0 )
		return A;

	// shift by the rest dist
	{
		vectorBoolInternalType bits1 { };
		vectorBoolInternalType bits2 { };

		// it might be more clear to iterate from the begining. However this is working nicely
		auto itAReverse = A.end ( ) - 1;

		// upper part of the last word in the vector can contain some garbage so it needs to be cleared
		vectorBoolInternalType maskTopWord = getMask ( sizeWithin );
		if ( maskTopWord != 0 )
			* ( itAReverse._M_p ) &= maskTopWord;

		// simulate behavior of reverse iterator
		while ( itAReverse >= A.begin ( ) ) {
			bits2 = * ( itAReverse._M_p );
			* ( itAReverse._M_p ) = * ( itAReverse._M_p ) >> distWithin | bits1 << backDist;
			bits1 = bits2;

			itAReverse._M_p --;
		}
	}

	return A;
}

template < class ... Ts >
vector < bool, Ts ... > operator >> ( vector < bool, Ts ... > A, size_t dist ) {
	A >>= dist;
	return A;
}

} /* namespace std */

#endif /* __VECTOR_HPP_ */
